""" PoP -- Prolog on Python  v0.1
    This is a library of Prolog engine.
"""

_version = '0.1'

import sys
import traceback

# exceptions
class Fail(Exception): pass

# types
TYPE_nil = 'nil'
TYPE_integer = 'integer'
TYPE_float = 'float'
TYPE_string = 'string'
TYPE_atom = 'atom'
TYPE_list = 'list'
TYPE_clause = 'clause'
TYPE_variable = 'variable'

# properties
PROP_standard_out = 'StandardOut'
PROP_on_arithmetic_error = 'OnArithmeticError'
PROP_trace_out = 'TraceOut'

# error messages
_ERR_name_not_string =  'Name is not string.'
_ERR_null_name = 'Name is null.'
_ERR_unknown_obj = 'Argument is unknown object. -- %s'
_ERR_invalid_type = 'Invalid type. -- %s'
_ERR_arg_not_sequence = 'Arguments are not in sequence.'
_ERR_body_not_sequence_nor_callable = 'Body is not sequence nor callable.'
_ERR_cond_not_sequence = 'Condition is not sequence.'
_ERR_empty_list = 'Empty list.'
_ERR_empty_body = 'Empty body.'
_ERR_empty_cond = 'Empty condition.'
_ERR_neither_rule_nor_fact = 'Neither rule nor fact.'
_ERR_invalid_keyword = 'Invalid keyword. -- %s'
_ERR_key_duplicated = 'Key duplicated.'
_ERR_invalid_arg_combination = 'Invalid argument combination.'
_ERR_invalid_value = 'Invalid value. -- %s'
_ERR_invalid_property = 'Invalid property. -- %s'

class _VList:
	def __init__(self, var=None):
		if var:
			self._vars = [var]
		else:
			self._vars = []
	def __len__(self):
		return len(self._vars)
	def __repr__(self):
		return `self._vars`
	def union(self, other):
		for x in other._vars:
			if not x in self._vars:
				self._vars.append(x)
	def setup_vdict(self):
		return _VDict(self._vars)
	
class _VDict:
	def __init__(self, vars):
		self._dict = {}
		for x in vars:
			if x:
				self._dict[x] = _Holder()
	def __repr__(self):
		d = self._dict.keys()
		d.sort()
		r = '{'
		p = ''
		for x in d:
			r = r + p + `x` + ':' + `self.active_holder(x)`
			p = ', '
		return r + '}'
	def active_holder(self, key):
		h = self._dict[key]
		while h.bound() and isinstance(h.item(), _Holder):
			h = h.item()
		return h
	def reconstruct(self):
		d = {}
		for x in self._dict.keys():
			h = self.active_holder(x)
			if h.bound():
				d[x] = h.item().reconstruct(h.vdict())
			else:
				d[x] = Variable('%s#%x' % (x, id(h)))
		return d

class _Cell:
	def __init__(self, item=None, next=None):
		self._item = item
		self._next = next
	def item(self):
		return self._item
	def set_item(self, item):
		self._item = item
	def next(self):
		return self._next
	def set_next(self, next):
		self._next = next

class _Seq:
	def __init__(self):
		self._first = self._last = None
	def __repr__(self):
		r = '['
		p = ''
		c = self._first
		while isinstance(c, _Cell):
			r = r + p + `c._item`
			c = c._next
			p = ', '
		return r + ']'
	def first(self):
		return self._first
	def last(self):
		return self._last
	def create_cursor(self):
		return _Cursor(self)
	def append(self, item):
		c = _Cell(item)
		if self._last:
			self._last.set_next(c)
			self._last = c
		else:
			self._first = self._last = c

class _Cursor:
	def __init__(self, list):
		self._list = list
		self._ptr = list.first()
	def item(self):
		if not isinstance(self._ptr, _Cell):
			raise EOFError()
		return self._ptr.item()
	def next(self):
		if not self._ptr or self._ptr == self._list.last():
			raise EOFError()
		self._ptr = self._ptr.next()
	def invalidate(self):
		self._ptr = None
	def rewind(self):
		self._ptr = self._list.first()

class _Holder:
	def __init__(self):
		pass
	def __repr__(self):
		try:
			return `self._item`
		except AttributeError:
			return '#%x' % id(self)
	def bind(self, item, vdict, binding):
		if item.includes_holder(self, vdict):
			raise Fail()
		self._item = item
		self._vdict = vdict
		binding.add(self)
	def unbind(self, binding):
		del self._item
		del self._vdict
		binding.remove(self)
	def bound(self):
		return self.__dict__.has_key('_item')
	def item(self):
		return self._item
	def vdict(self):
		return self._vdict
	def includes_holder(self, holder, vdict):
		return holder == self

class _Binding:
	def __init__(self):
		self._binding = []
	def __repr__(self):
		return `self._binding`
	def add(self, holder):
		self._binding.append(holder)
	def remove(self, holder):
		self._binding.remove(holder)
	def unbind_all(self):
		b = self._binding[:]
		for x in b:
			x.unbind(self)
		
class _Term:
	# root class
	def __init__(self):
		pass
	def type(self):
		return self._type

class _NonVar(_Term):
	def __init__(self):
		_Term.__init__(self)
	# def is_var(self, vdict):
		# return 0
	def compute(self, vdict):
		raise Fail()

class _BasicTerm(_NonVar):
	def __init__(self):
		_NonVar.__init__(self)
	def vlist(self):
		return _VList()
	def unify(self, vdict, other, ovdict, binding):
		if isinstance(other, Variable):
			other.unify(ovdict, self, vdict, binding)
		elif isinstance(other, self.__class__):
			pass
		else:
			raise Fail()
	def reconstruct(self, vdict):
		return self
	def includes_holder(self, holder, vdict):
		return 0

class _Nil(_BasicTerm):
	def __init__(self):
		_BasicTerm.__init__(self)
		self._type = TYPE_nil
	def __repr__(self):
		return '[]'

NIL = _Nil()

class _WrapperTerm(_BasicTerm):
	def __init__(self, val):
		_BasicTerm.__init__(self)
		self._val = val
	def __repr__(self):
		return `self._val`
	def value(self):
		return self._val
	def compute(self, vdict):
		return self
	def unify(self, vdict, other, ovdict, binding):
		if isinstance(other, Variable):
			other.unify(ovdict, self, vdict, binding)
		elif isinstance(other, self.__class__) and other._val == self._val:
			pass
		else:
			raise Fail()

class _Number(_WrapperTerm):
	def __init__(self, val):
		_WrapperTerm.__init__(self, val)
		self._val = val
	def unify(self, vdict, other, ovdict, binding):
		if isinstance(other, Variable):
			other.unify(ovdict, self, vdict, binding)
		elif isinstance(other, _Number) and other._val == self._val:
			pass
		else:
			raise Fail()

class Integer(_Number):
	"""integer :
	Integer(value)"""
	def __init__(self, val):
		if type(val) == type(0) or type(val) == type(0.0):
			_Number.__init__(self, int(val))
		else:
			raise TypeError(_ERR_invalid_type % `val`)
		self._type = TYPE_integer
 
class Float(_Number):
	"""float :
	Float(value)"""
	def __init__(self, val):
		if type(val) == type(0.0) or type(val) == type(0):
			_Number.__init__(self, float(val))
		else:
			raise TypeError(_ERR_invalid_type % `val`)
		self._type = TYPE_float
 
class String(_WrapperTerm):
	"""string :
	String(value)"""
	def __init__(self, val):
		if type(val) == type(''):
			_WrapperTerm.__init__(self, val)
		else:
			raise TypeError(_ERR_invalid_type % `val`)
		self._type = TYPE_string
	def __str__(self):
		return self._val

class _ConstructedTerm(_NonVar):
	def __init__(self):
		_NonVar.__init__(self)

class Atom(_ConstructedTerm):
	"""atom :
	Atom(name,x1,x2, .. ,xi,xj, .. ,xn) where n>=0
	Atom(name,x1,x2, .. ,xi,args=(xj, .. ,xn))
	Atom(name,x1,x2, .. ,xi,args=[xj, .. ,xn])
 <tips>
	a = Atom('a')
	a(1) is equivalent to Atom('a',1)"""
	def __init__(self, name, *nkw, **kw):
		_ConstructedTerm.__init__(self)
		self._type = TYPE_atom
		if type(name) <> type(''):
			raise TypeError(_ERR_name_not_string)
		if not name:
			raise ValueError(_ERR_null_name)
		self._name = name
		args = list(nkw)
		try:
			args = args + list(kw['args'])
			del kw['args']
		except KeyError:
			pass
		except TypeError:
			raise TypeError(_ERR_arg_not_sequence)
		if kw.keys():
			raise SyntaxError(_ERR_invalid_keyword % `kw.keys()`)
		self._args = []
		for x in args:
			try:
				self._args.append(_to_Term(x))
			except TypeError:
				raise
		if _functions.has_key(name):
			(argc, fn) = _functions[name]
			if len(args) == argc:
				self._fn = fn
			else:
				self._fn = None
		else:
			self._fn = None
	def __call__(self, *nkw, **kw):
		return apply(Atom, (self._name,)+tuple(self._args)+nkw, kw)
	def __repr__(self):
		r = self._name
		if self._args:
			r = self._name + '('
			p = ''
			for x in self._args:
				r = r + p + `x`
				p = ', '
			r = r + ')'
		return r
	def name(self):
		return self._name
	def args(self):
		return self._args
	def vlist(self):
		v = _VList()
		for x in self._args:
			v.union(x.vlist())
		return v
	def unify(self, vdict, other, ovdict, binding):
		if isinstance(other, Variable):
			other.unify(ovdict, self, vdict, binding)
		elif isinstance(other, Atom) and other._name == self._name and len(other._args) == len(self._args):
			for i in xrange(len(self._args)):
				self._args[i].unify(vdict, other._args[i], ovdict, binding)
		else:
			raise Fail()
	def reconstruct(self, vdict):
		if self._args:
			t = Atom(self._name, args=self._args[:])
			for i in xrange(len(t._args)):
				t._args[i] = t._args[i].reconstruct(vdict)
			return t
		else:
			return self
	def includes_holder(self, holder, vdict):
		for x in self._args:
			if x.includes_holder(holder, vdict):
				return 1
		return 0
	def compute(self, vdict):
		if not self._fn:
			raise Fail()
		a = ()
		for x in self._args:
			a = a + (x.compute(vdict),)
		return apply(self._fn, a)

class List(_ConstructedTerm, _Cell):
	"""list :
	List(x1,x2, .. ,xi,xj, .. ,xn) where n>=1
	List(x1,x2, .. ,xi,items=(xj, .. ,xn))
	List(x1,x2, .. ,xi,items=[xj, .. ,xn])
	List(x1,x2, .. ,xi,xj, .. ,xn,tail=y) where y is NIL, List or Variable
	List(x1,x2, .. ,xi,items=(xj, .. ,xn),tail=y)
	List(x1,x2, .. ,xi,items=[xj, .. ,xn],tail=y)"""
	def __init__(self, *nkw, **kw):
		_ConstructedTerm.__init__(self)
		_Cell.__init__(self)
		self._type = TYPE_list
		items = list(nkw)
		try:
			items = items + list(kw['items'])
			del kw['items']
		except KeyError:
			pass
		except TypeError:
			raise TypeError(_ERR_arg_not_sequence)
		try:
			tail = _to_Term(kw['tail'])
			del kw['tail']
		except KeyError:
			tail = NIL
		except TypeError:
			raise
		if kw.keys():
			raise SyntaxError(_ERR_invalid_keyword % `kw.keys()`)
		if not items:
			raise ValueError(_ERR_empty_list)
		if isinstance(tail, _Nil) or isinstance(tail, List) or isinstance(tail, Variable) :
			pass
		else:
			raise TypeError(_ERR_invalid_type % `tail`)
		try:
			self._item = _to_Term(items[0])
		except TypeError:
			raise
		if len(items) > 1:
			self._next = List(items=items[1:], tail=tail)
		else:
			self._next = tail
	def __repr__(self):
		c = self
		r = '['
		p = ''
		while isinstance(c, List):
			r = r + p + `c._item`
			c = c._next
			p = ', '
		if not isinstance(c, _Nil):
			r = r + ' | ' + `c`
		return r + ']'
	def vlist(self):
		v = _VList()
		c = self
		while isinstance(c, List):
			v.union(c._item.vlist())
			c = c._next
		v.union(c.vlist())
		return v
	def unify(self, vdict, other, ovdict, binding):
		if isinstance(other, Variable):
			other.unify(ovdict, self, vdict, binding)
		elif isinstance(other, List):
			self._item.unify(vdict, other._item, ovdict, binding)
			self._next.unify(vdict, other._next, ovdict, binding)
		else:
			raise Fail()
	def reconstruct(self, vdict):
		return List(self._item.reconstruct(vdict), tail=self._next.reconstruct(vdict))
	def includes_holder(self, holder, vdict):
		return self._item.includes_holder(holder, vdict) or self._next.includes_holder(holder, vdict)

class Clause(_ConstructedTerm):
	"""clause :
	Clause(head,x1,x2, .. ,xi,xj, .. ,xn) where n>=0
	Clause(head,x1,x2, .. ,xi,body=(xj, .. ,xn))
	Clause(head,x1,x2, .. ,xi,body=[xj, .. ,xn])
	Clause(head,body=callable_obj)"""
	def __init__(self, head, *nkw, **kw):
		_ConstructedTerm.__init__(self)
		self._type = TYPE_clause
		self._vlist = _VList()
		self._callback = None
		if head == None or isinstance(head, Atom):
			pass
		else:
			raise TypeError(_ERR_invalid_type % `head`)
		body = list(nkw)
		try:
			b = kw['body']
			del kw['body']
		except KeyError:
			b = []
		if isinstance(b, _Term):
			raise TypeError(_ERR_body_not_sequence_nor_callable)
		elif callable(b):
			if body:
				raise SyntaxError(_ERR_invalid_arg_combination)
			else:
				body = b
		else:
			try:
				body = body + list(b)
			except TypeError:
				raise TypeError(_ERR_body_not_sequence_nor_callable)
		if kw.keys():
			raise SyntaxError(_ERR_invalid_keyword % `kw.keys()`)
		if not head and not body:
			raise ValueError(_ERR_empty_body)
		if head:
			self._vlist.union(head.vlist())
		self._head = head
		if type(body) == type([]):
			self._body = []
			for x in body:
				try:
					xp = _to_Term(x)
				except TypeError:
					raise
				self._body.append(xp)
				self._vlist.union(xp.vlist())
		else:
			self._body = body
	def __repr__(self):
		r = ''
		if self._head:
			r = r + `self._head`
			if self._body:
				r = r + ' :- '
		else:
			r = r + '?- '
		if type(self._body) == type([]):
			p = ''
			for t in self._body:
				r = r + p + `t`
				p = ', '
		else:
			r = r + '*' + `self._body`
		return r
	def head(self):
		return self._head
	def body(self):
		return self._body
	def vlist(self):
		return self._vlist
	def is_query(self):
		return not self._head
	def is_fact(self):
		return not self._body
	def is_rule(self):
		return self._head and self._body
	def is_rule_or_fact(self):
		return self._head
	def unify(self, vdict, other, ovdict, binding):
		if isinstance(other, Variable):
			other.unify(ovdict, self, vdict, binding)
		elif isinstance(other, Clause):
			if type(self._body) == type([]):
				b = self._body
			else:
				b = []
			if type(other._body) == type([]):
				ob = other._body
			else:
				ob = []
			if len(b) == len(ob):
				self._head.unify(vdict, other, ovdict, binding)
				for i in xrange(len(b)):
					b[i].unify(vdict, ob[i], ovdict, binding)
			else:
				raise Fail()
		else:
			raise Fail()
	def reconstruct(self, vdict):
		if type(self._body) == type([]):
			c = Clause(self._head.reconstruct(vdict), body=self._body[:])
			for i in xrange(len(c._body)):
				c._body[i] = c._body[i].reconstruct(vdict)
		else:
			c = Clause(self._head.reconstruct(vdict), body=self._body)
		return c
	def includes_holder(self, holder, vdict):
		if self._head.includes_holder(holder, vdict):
			return 1
		if type(self._body) == type([]):
			for x in self._body:
				if x.includes_holder(holder, vdict):
					return 1
		return 0

class Variable(_Term):
	"""variable :
	Variable(name)
	Variable()  # anonymous variable"""
	def __init__(self, name=''):
		_Term.__init__(self)
		self._type = TYPE_variable
		if type(name) <> type(''):
			raise TypeError(_ERR_name_not_string)
		self._name = name
	def __repr__(self):
		return '_' + self._name
	def name(self):
		return self._name
	def vlist(self):
		return _VList(self._name)
	def holder(self, vdict):
		if self._name:
			return vdict.active_holder(self._name)
		else:
			return _Holder()
	# def is_var(self, vdict):
		# return not self.holder(vdict).bound()
	def unify(self, vdict, other, ovdict, binding):
		h = self.holder(vdict)
		if h.bound():
			h.item().unify(h.vdict(), other, ovdict, binding)
		elif isinstance(other, Variable):
			oh = other.holder(ovdict)
			if h == oh:
				pass
			elif oh.bound():
				h.bind(oh.item(), oh.vdict(), binding)
			else:
				h.bind(oh, ovdict, binding)
		else:
			h.bind(other, ovdict, binding)
	def reconstruct(self, vdict):
		h = self.holder(vdict)
		if h.bound():
			return h.item().reconstruct(h.vdict())
		elif self._name:
			return Variable('%s#%x' % (self._name, id(h)))
		else:
			return self
	def includes_holder(self, holder, vdict):
		h = self.holder(vdict)
		if h == holder:
			return 1
		elif h.bound():
			return h.item().includes_holder(holder, h.vdict())
		else:
			return 0
	def compute(self, vdict):
		h = self.holder(vdict)
		if h.bound():
			return h.item().compute(h.vdict())
		else:
			raise Fail()

class _BodyEval:
	def __init__(self, body, head, vdict, pop):
		self._body = body
		self._head = head
		self._vdict = vdict
		self._pop = pop
		self._proc = []
	def __repr__(self):
		return `self._body` + `self._vdict` + `self._proc`
	def head(self):
		return self._head
	def body(self):
		return self._body
	def vdict(self):
		return self._vdict
	def forward(self):
		if len(self._proc) >= len(self._body):
			return None
		else:
			p = _PredEval(self._body[len(self._proc)], self, self._vdict, self._pop)
			self._proc.append(p)
			return p
	def backward(self):
			del self._proc[-1]
	def cut(self):
		for p in self._proc:
			p.cut()
	def reconstruct(self):
		b = []
		for x in self._body:
			b.append(x.reconstruct(self._vdict))
		return b

class _PredEval:
	def __init__(self, pred, initiator, vdict, pop):
		self._eval_id = pop.generate_eval_id()
		if isinstance(pred, Variable):
			h = pred.holder(vdict)
			if h.bound():
				p, v = h.item(), h.vdict()
			else:
				p = v = None
		else:
			p, v  = pred, vdict
		if isinstance(p, Atom):
			self._pred, self._vdict = p, v
		else:
			self._pred = self._vdict = None
		self._initiator = initiator
		self._pop = pop
		self._binding = _Binding()
		self._success = 0
		self._cbmgr = None
	def __repr__(self):
		return `self._pred` + `self._vdict`
	def predicate(self):
		return self._pred
	def initiator(self):
		return self._initiator
	def evaluate(self, traceout):
		if traceout:
			self._trace1(traceout)
		try:
			if not self._pred:
				raise Fail()
			try:
				self._alt = self._pop.db().user(self._pred.name()).create_cursor()
			except KeyError:
				try:
					self._alt = self._pop.db().builtin(self._pred.name()).create_cursor()
				except KeyError:
					raise Fail()
			return self._evaluate(traceout)
		except Fail:
			if traceout:
				self._trace3(traceout)
			raise Fail()
	def reevaluate(self, traceout):
		self._exec_cancel(traceout)
		self._binding.unbind_all()
		if traceout:
			self._trace1(traceout)
		try:
			try:
				self._alt.next()
			except EOFError:
				raise Fail()
			return self._evaluate(traceout)
		except Fail:
			if traceout:
				self._trace3(traceout)
			raise Fail()
	def _evaluate(self, traceout):
		while 1:
			c = self._alt.item()
			w = c.vlist().setup_vdict()
			try:
				self._pred.unify(self._vdict, c.head(), w, self._binding)
			except Fail:
				self._binding.unbind_all()
				try:
					self._alt.next()
				except EOFError:
					raise Fail()
				continue
			b = c.body()
			if type(b) == type([]):
				if b:
					be = _BodyEval(c.body(), self, w, self._pop)
				else:
					be = None
				break
			else:
				try:
					r = self._exec_call(b, c, w, traceout)
				except Fail:
					r = 0
				if r:
					be = None
					break
				else:
					self._exec_cancel(traceout)
					self._binding.unbind_all()
					try:
						self._alt.next()
					except EOFError:
						raise Fail()
					continue
		self._success = self._success + 1
		if traceout:
			self._trace2(traceout, c, w)
		return be
	def _exec_call(self, fn, myclause, vdict, traceout):
		try:
			if isinstance(fn, _BuiltinCallback):
				return self._exec_builtin(fn, myclause, vdict, traceout)
			else:
				self._cbmgr = _CallbackManager(fn, vdict, self._eval_id)
				return self._cbmgr.exec_call(traceout)
		except Fail:
			raise
		except ArithmeticError:
			if self._pop.properties()[PROP_on_arithmetic_error] == 'Fail':
				raise Fail()
			else:
				raise
		except:
			raise Fail()
	def _exec_cancel(self, traceout):
		if self._cbmgr:
			self._cbmgr.exec_cancel(traceout)
	def _exec_builtin(self, fn, myclause, vdict, traceout):
		if traceout:
			traceout.write(' [spy]  trying %s ...\n' % `fn`)
		try:
			return fn(myclause, self, vdict, self._binding)
		except:
			if traceout:
				traceout.write(' [spy]  exception raised -- as follows...\n')
				traceback.print_exc(file=traceout.file())
			raise
	def cleanup(self, traceout):
		if self._cbmgr:
			self._cbmgr.cleanup(traceout)
	def cut(self):
		self._alt.invalidate()
	def rewind(self):
		self._alt.rewind()
	def reconstruct(self):
		return self._pred.reconstruct(self._vdict)
	def _trace1(self, traceout):
		p = self.reconstruct()
		if self._success == 0:
			traceout.write(' [spy] C<%d,%d> %s\n' % (self._eval_id, self._success, `p`))
		else:
			traceout.write(' [spy] R<%d,%d> %s\n' % (self._eval_id, self._success, `p`))
	def _trace2(self, traceout, clause, vdict):
		h = clause.head()
		b = clause.body()
		c = clause.reconstruct(vdict)
		if type(b) == type([]):
			if b:
				traceout.write(' [spy] => %s --> %s\n' % (`h`, `c`))
			else:
				traceout.write(' [spy] >> %s --> %s\n' % (`h`, `c`))
		else:
			traceout.write(' [spy] >> %s <-> %s\n' % (`h`, `c`))
	def _trace3(self, traceout):
		traceout.write(' [spy] << FAIL\n')

class _BuiltinCallback:
	def __init__(self, name, fn):
		self._name = name
		self._fn = fn
	def __repr__(self):
		return self._name
	def __call__(self, *args):
		return apply(self._fn, args)

class _CallbackManager:
	def __init__(self, fn, vdict, eval_id):
		self._fn = fn
		self._vdict = vdict
		self._eval_id = eval_id
		self._binding = _Binding()
		self._cleaner = self._cleaner_args = None
	def fn(self):
		return self._fn
	def exec_call(self, traceout):
		self._calling = 1
		self._traceout = traceout
		if traceout:
			traceout.write(' [spy]  trying %s ...\n' % `self._fn`)
		try:
			return self._fn(self)
		except:
			if traceout:
				traceout.write(' [spy]  exception raised -- as follows...\n')
				traceback.print_exc(file=traceout.file())
			raise
	def exec_cancel(self, traceout):
		self._calling = 0
		self._traceout = traceout
		if traceout:
			traceout.write(' [spy]  canceling %s ...\n' % `self._fn`)
		try:
			self._fn(self)
		except:
			if traceout:
				traceout.write(' [spy]  exception raised (ignored) -- as follows...\n')
				traceback.print_exc(file=traceout.file())
			self._binding.unbind_all()
		else:
			self._binding.unbind_all()
	def cleanup(self, traceout):
		if self._cleaner:
			(nkw, kw) = self._cleaner_args
			if traceout:
				traceout.write(' [spy]  calling cleaner %s ...\n' % `self._cleaner`)
			try:
				apply(self._cleaner, nkw, kw)
			except:
				if traceout:
					traceout.write(' [spy]  exception raised (ignored) -- as follows...\n')
					traceback.print_exc(file=traceout.file())
	def eval_id(self):
		return self._eval_id
	def calling(self):
		return self._calling
	def variables(self):
		return self._vdict.reconstruct()
	def traceout(self):
		if self._traceout:
			return self._traceout.file()
		else:
			return None
	def bind(self, var, item):
		if not self._calling:
			raise Fail()
		h = self._vdict.active_holder(var)
		if h.bound():
			raise Fail()
		t = _to_Term(item)
		if t.vlist():
			raise Fail()
		h.bind(t, self._vdict, self._binding)
	def register_cleaner(self, fn, nkwargs=(), kwargs={}):
		if not callable(fn) or isinstance(fn, _Term):
			raise TypeError(_ERR_invalid_type % `fn`)
		if type(nkwargs) <> type(()):
			raise TypeError(_ERR_invalid_type % `nkwargs`)
		if type(kwargs) <> type({}):
			raise TypeError(_ERR_invalid_type % `kwargs`)
		self._cleaner = fn
		self._cleaner_args = (nkwargs, kwargs)

class _RuleDB:
	def __init__(self):
		self._udict = {}
		self._bdict = {}
	def __repr__(self):
		r = ''
		for p in self._udict.keys():
			r = r + `self._udict[p]`
		return r
	def insert(self, clause):
		p = clause.head().name()
		if not self._udict.has_key(p):
			self._udict[p] = _Seq()
		self._udict[p].append(clause)
	def delete(self, pred):
		del self._udict[pred]
	def insert_builtin(self, clause):
		p = clause.head().name()
		if not self._bdict.has_key(p):
			self._bdict[p] = _Seq()
		self._bdict[p].append(clause)
	def user(self, name):
		return self._udict[name]
	def builtin(self, name):
		return self._bdict[name]
	def userdefs(self):
		p = self._udict.keys()
		p.sort()
		return p
	def builtins(self):
		p = self._bdict.keys()
		p.sort()
		return p

class _OutStream:
	def __init__(self, file=sys.stdout):
		self._file = file
	def file(self):
		return self._file
	def set_file(self, file):
		self._file = file
	def write(self, stream):
		self._file.write(stream)
			
def _to_Term(x):
	if isinstance(x, _Term):
		return x
	elif type(x) == type(0):
		return Integer(x)
	elif type(x) == type(0.0):
		return Float(x)
	elif type(x) == type(''):
		return String(x)
	elif x == []:
		return NIL
	elif type(x) == type([]):
		return List(items=x)
	else:
		raise TypeError(_ERR_unknown_obj % `x`)

class PoP:
	version = _version
	def __init__(self, prop={}):
		self._stdout = _OutStream(sys.stdout)
		self._traceout = self._stdout
		self._db = _RuleDB()
		self._query = None
		self._eval = None
		self._proc = []
		self._eval_id = 0
		self._result = None
		self._spy = {}
		self._trcmask = self._trcmaskc = None
		self._setup_builtins()
		self._prop = {
			PROP_standard_out: sys.stdout,
			PROP_trace_out: PROP_standard_out,
			PROP_on_arithmetic_error: 'Abort'
		}
		for (k, v) in prop.items():
			self.change_property(k, v)
	def db(self):
		return self._db
	def generate_eval_id(self):
		i = self._eval_id
		self._eval_id = self._eval_id + 1
		return i
	def last_query(self):
		return self._query
	def last_result(self):
		return self._result
	def userdefs(self):
		return self._db.userdefs()
	def builtins(self):
		return self._db.builtins()
	def userpred(self, name):
		try:
			c = self._db.user(name).create_cursor()
		except KeyError:
			return []
		p = []
		while 1:
			try:
				p.append(c.item())
				c.next()
			except EOFError:
				break
		return p
	def builtinpred(self, name):
		try:
			c = self._db.builtin(name).create_cursor()
		except KeyError:
			return []
		p = []
		while 1:
			try:
				p.append(c.item())
				c.next()
			except EOFError:
				break
		return p
	def functions(self):
		f = _functions.keys()
		f.sort()
		return f
	def listing(self, name='.*'):
		import re
		n = re.compile(name)
		for x in self.userdefs():
			if n.match(x):
				for y in self.userpred(x):
					self._stdout.write(`y`+'.\n')
	def assertz(self, *nkw, **kw):
		try:
			c = kw['clause']
			del kw['clause']
			if nkw:
				raise SyntaxError(_ERR_invalid_arg_combination)
			if kw:
				raise SyntaxError(_ERR_invalid_keyword % `kw.keys()`)
			if not isinstance(c, Clause):
				raise TypeError(_ERR_invalid_type % `c`)
		except KeyError:
			c = apply(Clause, nkw, kw)
		if not c.is_rule_or_fact():
			raise ValueError(_ERR_neither_rule_nor_fact)
		self._db.insert(c)
	def abolish(self, name):
		self._db.delete(name)
	def spy(self, name, key=None):
		if key == None:
			k = name
		else:
			k = key
		if type(k) <> type(''):
			raise TypeError(_ERR_invalid_type % `k`)
		self._spy[name] = k
	def nospy(self, name):
		del self._spy[name]
	def spies(self):
		return self._spy
	def trace(self, mask='.*'):
		import re
		self._trcmask = mask
		self._trcmaskc = re.compile(mask)
	def notrace(self):
		self._trcmask = self._trcmaskc = None
	def trace_mask(self):
		return self._trcmask
	def inquire(self, *nkw, **kw):
		cond = list(nkw)
		try:
			cond = cond + kw['cond']
			del kw['cond']
		except KeyError:
			pass
		except TypeError:
			raise TypeError(_ERR_cond_not_sequence)
		if not cond:
			raise ValueError(_ERR_empty_cond)
		if kw.keys():
			raise SyntaxError(_ERR_invalid_keyword % `kw.keys()`)
		self._query = Clause(None, body=cond)
		self._eval = _BodyEval(self._query.body(), None, self._query.vlist().setup_vdict(), self)
		self._proc = []
		self._eval_id = 0
		try:
			self._evaluate('F')
			self._result = self._eval.vdict().reconstruct()
		except Fail:
			self._result = None
		return self._result
	def inquire_next(self):
		if not self._proc:
			return None
		try:
			self._evaluate('r')
			self._result = self._eval.vdict().reconstruct()
		except Fail:
			self._result = None
		return self._result
	def _evaluate(self, action):
		act = action
		b = self._eval
		while 1:
			if act == 'F':
				p = b.forward()
				if p:
					self._proc.append(p)
					act = 'e'
				else:
					p = b.head()
					if not p:
						break
					b = p.initiator()
					act = 'F'
				continue
			if act == 'B':
				b.backward()
				del self._proc[-1]
				if not self._proc:
					raise Fail()
				act = 'r'
				continue
			if act == 'e' or act == 'r':
				p = self._proc[-1]
				if self._trcmaskc:
					try:
						k = self._spy[p.predicate().name()]
						if self._trcmaskc.match(k):
							traceout = self._traceout
						else:
							traceout = None
					except KeyError:
						traceout = None
				else:
					traceout = None
				try:
					if act == 'e':
						b = p.evaluate(traceout)
					else:
						b = p.reevaluate(traceout)
				except Fail:
					b = p.initiator()
					p.cleanup(traceout)
					act = 'B'
					continue
				if not b:
					b = p.initiator()
				act = 'F'
				continue
			assert 0, 'Never reached.'
	def properties(self):
		return self._prop.copy()
	def change_property(self, key, val):
		if key == PROP_standard_out:
			self._prop[PROP_standard_out] = val
			self._stdout.set_file(val)
		elif key == PROP_trace_out:
			if val == PROP_standard_out:
				self._prop[PROP_trace_out] = val
				self._traceout = self._stdout
			else:
				self._prop[PROP_trace_out] = val
				self._traceout = _OutStream(val)
		elif key == PROP_on_arithmetic_error:
			if val == 'Fail' or val == 'Abort':
				self._prop[PROP_on_arithmetic_error] = val
			else:
				raise ValueError(_ERR_invalid_value % `val`)
		else:
			raise KeyError(_ERR_invalid_property % `key`)

	# ========== builtin predicates ==========
	def _builtin_cut(self, myclause, predeval, vdict, binding):
		predeval.initiator().cut()
		predeval.initiator().head().cut()
		return 1
	def _builtin_repeat(self, myclause, predeval, vdict, binding):
		predeval.rewind()
		return 1
	def _builtin_var(self, myclause, predeval, vdict, binding):
		return not vdict.active_holder('v1').bound()
	def _builtin_nonvar(self, myclause, predeval, vdict, binding):
		return vdict.active_holder('v1').bound()
	def _builtin_integer(self, myclause, predeval, vdict, binding):
		h1 = vdict.active_holder('v1')
		return h1.bound() and isinstance(h1.item(), Integer)
	def _builtin_float(self, myclause, predeval, vdict, binding):
		h1 = vdict.active_holder('v1')
		return h1.bound() and isinstance(h1.item(), Float)
	def _builtin_number(self, myclause, predeval, vdict, binding):
		h1 = vdict.active_holder('v1')
		return h1.bound() and isinstance(h1.item(), _Number)
	def _builtin_string(self, myclause, predeval, vdict, binding):
		h1 = vdict.active_holder('v1')
		return h1.bound() and isinstance(h1.item(), String)
	def _builtin_nil(self, myclause, predeval, vdict, binding):
		h1 = vdict.active_holder('v1')
		return h1.bound() and isinstance(h1.item(), _Nil)
	def _builtin_atom(self, myclause, predeval, vdict, binding):
		h1 = vdict.active_holder('v1')
		return h1.bound() and isinstance(h1.item(), Atom)
	def _builtin_atom1(self, myclause, predeval, vdict, binding):
		h1 = vdict.active_holder('v1')
		return h1.bound() and isinstance(h1.item(), Atom) and not h1.item().args()
	def _builtin_atom2(self, myclause, predeval, vdict, binding):
		h1 = vdict.active_holder('v1')
		return h1.bound() and isinstance(h1.item(), Atom) and h1.item().args()
	def _builtin_list(self, myclause, predeval, vdict, binding):
		h1 = vdict.active_holder('v1')
		return h1.bound() and isinstance(h1.item(), List)
	def _builtin_clause(self, myclause, predeval, vdict, binding):
		h1 = vdict.active_holder('v1')
		return h1.bound() and isinstance(h1.item(), Clause)
	def _builtin_is(self, myclause, predeval, vdict, binding):
		[x1, x2] = myclause.head().args()
		r = x2.compute(vdict)
		x1.unify(vdict, r, None, binding)
		return 1
	def _builtin_eq(self, myclause, predeval, vdict, binding):
		return self._cmp(myclause, predeval, vdict, binding) == 0
	def _builtin_ne(self, myclause, predeval, vdict, binding):
		return self._cmp(myclause, predeval, vdict, binding) <> 0
	def _builtin_le(self, myclause, predeval, vdict, binding):
		return self._cmp(myclause, predeval, vdict, binding) <= 0
	def _builtin_lt(self, myclause, predeval, vdict, binding):
		return self._cmp(myclause, predeval, vdict, binding) < 0
	def _builtin_ge(self, myclause, predeval, vdict, binding):
		return self._cmp(myclause, predeval, vdict, binding) >= 0
	def _builtin_gt(self, myclause, predeval, vdict, binding):
		return self._cmp(myclause, predeval, vdict, binding) > 0
	def _cmp(self, myclause, predeval, vdict, binding):
		[x1, x2] = myclause.head().args()
		y1 = x1.compute(vdict)
		y2 = x2.compute(vdict)
		if (isinstance(y1, _Number) and isinstance(y2, _Number)) or (isinstance(y1, String) and isinstance(y2, String)):
			return cmp(y1.value(), y2.value())
		else:
			raise Fail()
	def _builtin_write(self, myclause, predeval, vdict, binding):
		vd = vdict.reconstruct()
		self._stdout.write(str(vd['v1']))
		return 1
	# ---------- setup builtin predicates ----------
	def _setup_builtins(self):
		BC = _BuiltinCallback  # just an alias
		Cut_ = Atom('!')
		True_ = Atom('true')
		Fail_ = Atom('fail')
		Repeat_ = Atom('repeat')
		Not_ = Atom('not')
		Var_ = Atom('var')
		Nonvar_ = Atom('nonvar')
		Integer_ = Atom('integer')
		Float_ = Atom('float')
		Number_ = Atom('number')
		String_ = Atom('string')
		Nil_ = Atom('nil')
		Atom_ = Atom('atom')
		Atom1_ = Atom('atom1')
		Atom2_ = Atom('atom2')
		List_ = Atom('list')
		Clause_ = Atom('clause')
		Is_ = Atom('is')
		Eq_ = Atom('eq')
		Ne_ = Atom('ne')
		Le_ = Atom('le')
		Lt_ = Atom('lt')
		Ge_ = Atom('ge')
		Gt_ = Atom('gt')
		Equal_ = Atom('equal')
		Nequal_ = Atom('nequal')
		Write_ = Atom('write')
		V = Variable()
		V1 = Variable('v1')
		V2 = Variable('v2')
		builtins = [
			(Cut_, BC('builtin_cut', self._builtin_cut)),
			(True_, []),
			(Fail_, BC('builtin_fail', lambda *args: 0)),
			(Repeat_, []),
			(Repeat_, BC('builtin_repeat', self._builtin_repeat)),
			(Not_(V1), [V1, Cut_, Fail_]),
			(Not_(V), []),
			(Var_(V1), BC('builtin_var', self._builtin_var)),
			(Nonvar_(V1), BC('builtin_nonvar', self._builtin_nonvar)),
			(Integer_(V1), BC('builtin_integer', self._builtin_integer)),
			(Float_(V1), BC('builtin_float', self._builtin_float)),
			(Number_(V1), BC('builtin_number', self._builtin_number)),
			(String_(V1), BC('builtin_string', self._builtin_string)),
			(Nil_(V1), BC('builtin_nil', self._builtin_nil)),
			(Atom_(V1), BC('builtin_atom', self._builtin_atom)),
			(Atom1_(V1), BC('builtin_atom1', self._builtin_atom1)),
			(Atom2_(V1), BC('builtin_atom2', self._builtin_atom2)),
			(List_(V1), BC('builtin_list', self._builtin_list)),
			(Clause_(V1), BC('builtin_clause', self._builtin_clause)),
			(Is_(V1, V2), BC('builtin_is', self._builtin_is)),
			(Eq_(V1, V2), BC('builtin_eq', self._builtin_eq)),
			(Ne_(V1, V2), BC('builtin_ne', self._builtin_ne)),
			(Le_(V1, V2), BC('builtin_le', self._builtin_le)),
			(Lt_(V1, V2), BC('builtin_lt', self._builtin_lt)),
			(Ge_(V1, V2), BC('builtin_ge', self._builtin_ge)),
			(Gt_(V1, V2), BC('builtin_gt', self._builtin_gt)),
			(Equal_(V1, V1), []),
			(Nequal_(V1, V2), [Not_(Equal_(V1, V2))]),
			(Write_(V1), BC('builtin_write', self._builtin_write))
		]
		for (hd, bd) in builtins:
			self._db.insert_builtin(Clause(hd, body=bd))

# ========== functions ==========
def _func_integer(v1):
	if isinstance(v1, Integer):
		return v1
	elif isinstance(v1, Float):
		return Integer(v1.value())
	else:
		raise Fail()
def _func_float(v1):
	if isinstance(v1, Float):
		return v1
	elif isinstance(v1, Integer):
		return Float(v1.value())
	else:
		raise Fail()
def _func_abs(v1):
	if isinstance(v1, _Number):
		x1 = v1.value()
		if x1 >= 0:
			return v1
		else:
			return _to_Term(-x1)
	else:
		raise Fail()
def _func_add(v1, v2):
	if isinstance(v1, _Number) and isinstance(v2, _Number):
		return _to_Term(v1.value()+v2.value())
	else:
		raise Fail()
def _func_sub(v1, v2):
	if isinstance(v1, _Number) and isinstance(v2, _Number):
		return _to_Term(v1.value()-v2.value())
	else:
		raise Fail()
def _func_mul(v1, v2):
	if isinstance(v1, _Number) and isinstance(v2, _Number):
		return _to_Term(v1.value()*v2.value())
	else:
		raise Fail()
def _func_div(v1, v2):
	if isinstance(v1, _Number) and isinstance(v2, _Number):
		return _to_Term(float(v1.value())/float(v2.value()))
	else:
		raise Fail()
def _func_idiv(v1, v2):
	if isinstance(v1, Integer) and isinstance(v2, Integer):
		return _to_Term(v1.value()/v2.value())
	else:
		raise Fail()
def _func_mod(v1, v2):
	if isinstance(v1, Integer) and isinstance(v2, Integer):
		return _to_Term(v1.value()%v2.value())
	else:
		raise Fail()
# ---------- function table ----------
_functions = {
	'integer': (1, _func_integer),
	'float': (1, _func_float),
	'abs': (1, _func_abs),
	'add': (2, _func_add),
	'sub': (2, _func_sub),
	'mul': (2, _func_mul),
	'div': (2, _func_div),
	'idiv': (2, _func_idiv),
	'mod': (2, _func_mod)
}
